@@ -1,34 +1,20 @@
 import re
+import inspect
 from types import MethodType
-from django.urls import reverse
-# from bootstrap_modals.helper import show_modal
 from typing import TypeVar, Dict
-
+from .helpers import get_url, render_replace
 KT = TypeVar('KT')
 VT = TypeVar('VT')
 
 
-def row_button(command, button_text, *, function='Html', button_classes='btn btn-sm', **kwargs):
-    rb = {
-        'html': (f'<button data-command="{command}" onclick="django_datatables.b_r(this)" '
-                 f'class="{button_classes}">{button_text}</button>'),
-        'function': function,
-    }
-    rb.update(kwargs)
-    return rb
-
-
-def render_replace(*, column, var='%1%', html=None):
-    if html:
-        return {'var': var, 'column': column, 'html': html, 'function': 'Replace'}
-    else:
-        return {'var': var, 'column': column, 'function': 'Replace'}
-
-
 class ColumnNameError(Exception):
     pass
 
 
+class DatatableColumnError(Exception):
+    pass
+
+
 class ColumnBase:
 
     @staticmethod
@@ -72,7 +58,10 @@ class ColumnBase:
         self.column_defs = {}
         self.column_name, options = self.extract_options(kwargs.get('column_name', ''))
         self.options: Dict[KT, VT] = options
-        self.model_path = self.get_model_path(self.column_name)
+        self.model_path = kwargs.get('model_path')
+        if not self.model_path:
+            self.model_path = self.get_model_path(self.column_name)
+        self.model = kwargs.get('model')
         if kwargs.get('column_name') == field:
             self.field = self.column_name
         else:
@@ -111,7 +100,7 @@ class ColumnBase:
 
     @column_name.setter
     def column_name(self, value):
-        if not re.match('^[A-Za-z0-9._]+$', value):
+        if not re.match('^[A-Za-z0-9._/]+$', value):
             raise ColumnNameError('Invalid column_name: ' + value)
         self._column_name = value
 
@@ -163,14 +152,12 @@ class ColumnBase:
     def __row_result(self, data_dict, _page_results):
         if isinstance(self.field, (list, tuple)):
             return [data_dict.get(f) for f in self.field]
+        if self.options.get('choices'):
+            return self.options['choices'].get(data_dict.get(self.field))
         return data_dict.get(self.field)
 
     def style(self):
         col_def_str = self.column_defs
-#        if self.column_type == ColumnDef.COL_CURRENCY or self.column_type == ColumnDef.COL_CURRENCY_PENCE:
-#            colDefStr['className'] = 'dt-body-right pr-4'
-        if 'mobile' in self.options and not (self.options['mobile']):
-            col_def_str['mobile'] = False
         if self.options.get('hidden'):
             col_def_str['visible'] = False
             col_def_str['searchable'] = False
@@ -259,13 +246,7 @@ class ColumnLink(ColumnBase):
 
     @url.setter
     def url(self, url_name):
-        if type(url_name) == tuple:
-            self._url = reverse(url_name[0], args=[*url_name[1:]])
-        else:
-            if url_name.find('999999') == -1:
-                self._url = reverse(url_name, args=[999999])
-            else:
-                self._url = url_name
+        self._url = get_url(url_name)
 
     def __init__(self, *, url_name, link_ref_column='id', link_html='%1%', var='%1%', **kwargs):
         if not self.initialise(locals()):
@@ -285,27 +266,37 @@ class ColumnLink(ColumnBase):
 class ManyToManyColumn(DatatableColumn):
 
     def setup_results(self, request, all_results):
-        tags = self.m2m_model.objects.values_list(self.connecting_field + '__id', 'id')
+        if self.reverse:
+            tags = self.related_model.objects.values_list(self.field_id, 'id')
+        else:
+            tags = self.model.objects.values_list('id', self.field_id)
+        tags = tags.filter(**{self.field_id + '__isnull': False}).distinct()
         tag_dict = {}
         for t in tags:
             tag_dict.setdefault(t[0], []).append(t[1])
         all_results['m2m' + self.column_name] = tag_dict
 
-    def proc_result(self, data_dict, page_results):
+    def row_result(self, data_dict, page_results):
         return page_results['m2m' + self.column_name].get(data_dict['id'], [])
 
-    def __init__(self, *, m2m_model, connecting_field, display_field, html=' %1% ', **kwargs):
+    def __init__(self, *,  html=' %1% ', **kwargs):
         if not self.initialise(locals()):
             return
         super().__init__(**self.kwargs)
-        self.m2m_model = self.options.pop('m2m_model')
-        self.connecting_field = connecting_field
-        self.display_field = display_field
-        self.options['lookup'] = list(m2m_model.objects.values_list('id', display_field))
-        self.row_result = self.proc_result
-        self.options['render'] = [
-            {'var': '%1%', 'html': html, 'function': 'ReplaceLookup'},
-        ]
+        fields = self.field.split('__')
+        if not inspect.isclass(self.model):
+            raise DatatableColumnError('ManyToManyColumn must have model set')
+        connecting_field = self.model._meta.get_field(fields[-2])
+        self.related_model = connecting_field.related_model
+        if hasattr(connecting_field, 'field'):
+            self.field_id = connecting_field.field.attname + '__id'
+            self.reverse = True
+        else:
+            self.field_id = fields[-2] + '__id'
+            self.reverse = False
+        self.field = None
+        self.options['lookup'] = list(self.related_model.objects.values_list('id', fields[-1]))
+        self.options['render'] = [{'var': '%1%', 'html': html, 'function': 'ReplaceLookup'}]
 
 
 class DateColumn(ColumnBase):
